/*
Write-once filesystem

Copyright (c) 2022 Aron Barath
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the Author nor the names of contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Source code: svn://svn.repo.hu/wroncefs/trunk
Contact the author: aron-dev@mailbox.org
*/

#include "impl.h"
#include "wroncefs_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

static char line[1024];

static const char delims[] = " \t\r\n\v";

struct cmd
{
	const char* name;
	const char* help;
	void (*func)(const struct impl* const impl);
};

static const struct cmd cmds[];

static const struct cmd* find_cmd(const char* const name)
{
	for(const struct cmd* cmd=cmds;NULL!=cmd->name;++cmd)
	{
		if(0==strcmp(name, cmd->name))
		{
			return cmd;
		}
	}

	return NULL;
}

static void func_help(const struct impl* const impl)
{
	char* name = strtok(NULL, delims);

	if(NULL==name)
	{
		size_t n = 0;

		printf("Command list:\n");

		for(const struct cmd* cmd=cmds;NULL!=cmd->name;++n,++cmd)
		{
			printf("  %s - %s\n", cmd->name, cmd->help);
		}

		printf("%zu commands\n", n);
		return;
	}

	do
	{
		const struct cmd* const cmd = find_cmd(name);

		if(NULL!=cmd)
		{
			printf("'%s': %s\n", cmd->name, cmd->help);
		}
		else
		{
			printf("Command '%s' not found\n", name);
		}
	}
	while(NULL!=(name=strtok(NULL, delims)));
}

static void func_df(const struct impl* const impl)
{
	struct w_statfs st;

	const int ec = impl->f_statfs(&st);

	if(0!=ec)
	{
		printf("Could not stat filesystem: %s\n", impl->f_strerror(ec));
		return;
	}

	printf("mode:     %s,%s\n",
		(impl->inf_can_alter ? "alter" : "noalter"),
		(impl->inf_16bit_mode ? "16-bit" : "32-bit"));
	printf("filesize: %zu\n", (size_t)impl->max_filesize);
	printf("bsize:    %zu\n", st.bsize);
	printf("bfree:    %zu\n", st.bfree);
	printf("blocks:   %zu\n", st.blocks);
	printf("namelen:  %zu\n", st.namelen);
	printf("use:      %zu%%\n", (100*(st.blocks - st.bfree))/st.blocks);
	printf("1K blks:  %zu\n", (st.blocks*st.bsize)/1024);
	printf("1K free:  %zu\n", (st.bfree*st.bsize)/1024);
}

static void func_fsck(const struct impl* const impl)
{
	const int ec = impl->f_fsck();

	if(0==ec)
	{
		printf("Success\n");
	}
	else
	{
		printf("Error: %s\n", impl->f_strerror(ec));
	}
}

static void func_cd(const struct impl* const impl)
{
	char* const name = strtok(NULL, delims);

	if(NULL==name)
	{
		return;
	}

	const int ec = impl->f_chdir(name);

	if(0!=ec)
	{
		printf("Error: %s\n", impl->f_strerror(ec));
	}
}

static void func_mkdir(const struct impl* const impl)
{
	char* const name = strtok(NULL, delims);

	if(NULL==name)
	{
		printf("Missing argument\n");
		return;
	}

	const int ec = impl->f_mkdir(name);

	if(0!=ec)
	{
		printf("Error: %s\n", impl->f_strerror(ec));
	}
}

static void func_rmdir(const struct impl* const impl)
{
	char* const name = strtok(NULL, delims);

	if(NULL==name)
	{
		printf("Missing argument\n");
		return;
	}

	const int ec = impl->f_rmdir(name);

	if(0!=ec)
	{
		printf("Error: %s\n", impl->f_strerror(ec));
	}
}

static void func_unlink(const struct impl* const impl)
{
	char* const name = strtok(NULL, delims);

	if(NULL==name)
	{
		printf("Missing argument\n");
		return;
	}

	const int ec = impl->f_unlink(name);

	if(0!=ec)
	{
		printf("Error: %s\n", impl->f_strerror(ec));
	}
}

static void func_ls(const struct impl* const impl)
{
	const char* name = strtok(NULL, delims);

	if(NULL==name)
	{
		name = ".";
	}

	int ec;

	struct w_dir* const dir = impl->f_opendir(name, &ec);

	if(NULL==dir)
	{
		printf("Error: %s\n", impl->f_strerror(ec));
		return;
	}

	for(struct w_dirent* ent;NULL!=(ent=impl->f_readdir(dir, &ec));)
	{
		struct w_stat st;

		if(0!=(ec=impl->f_istat(ent->ino, &st)))
		{
			printf("Failed to stat '%s' (ino #%zu): %s\n", ent->name,
				ent->ino, impl->f_strerror(ec));
			continue;
		}

		printf("%6zu %10zu %s%s\n", ent->ino, st.size, ent->name,
			(st.is_file?"":"/"));
	}

	if(0!=ec)
	{
		printf("Error: %s\n", impl->f_strerror(ec));
	}

	impl->f_closedir(dir);
}

static void func_cat(const struct impl* const impl)
{
	const char* const name = strtok(NULL, delims);

	if(NULL==name)
	{
		printf("Error: missing argument\n");
		return;
	}

	int ec;

	struct w_file* const fd = impl->f_open(name, &ec);

	if(NULL==fd)
	{
		printf("Error: %s\n", impl->f_strerror(ec));
		return;
	}

	char buf[2048];
	int  size;

	while(0<(size=impl->f_read(fd, buf, sizeof(buf), &ec)))
	{
		fwrite(buf, size, 1, stdout);
		fflush(stdout);
	}

	if(0!=ec)
	{
		printf("\nRead error: %s\n", impl->f_strerror(ec));
	}

	impl->f_close(fd);
}

static void func_export(const struct impl* const impl)
{
	const char* const name  = strtok(NULL, delims);
	const char* const hname = strtok(NULL, delims);

	if(NULL==name || NULL==hname)
	{
		printf("Error: missing argument\n");
		return;
	}

	int ec;

	struct w_file* const fd = impl->f_open(name, &ec);

	if(NULL==fd)
	{
		printf("Error: %s\n", impl->f_strerror(ec));
		return;
	}

	FILE* const file = fopen(hname, "wb");

	if(NULL==file)
	{
		const int err = errno;
		fprintf(stderr, "Error: Could not create '%s': %s\n", hname,
			strerror(err));
		impl->f_close(fd);
		return;
	}

	char buf[2048];
	int  size;

	while(0<(size=impl->f_read(fd, buf, sizeof(buf), &ec)))
	{
		fwrite(buf, size, 1, file);
		fflush(stdout);
	}

	if(0!=ec)
	{
		printf("\nRead error: %s\n", impl->f_strerror(ec));
	}

	fclose(file);
	impl->f_close(fd);
}

static void func_import(const struct impl* const impl)
{
	const char* const name  = strtok(NULL, delims);
	const char* const hname = strtok(NULL, delims);

	if(NULL==name || NULL==hname)
	{
		printf("Error: missing argument\n");
		return;
	}

	int ec;

	struct w_file* const fd = impl->f_creat(name, &ec);

	if(NULL==fd)
	{
		printf("Error: %s\n", impl->f_strerror(ec));
		return;
	}

	FILE* const file = fopen(hname, "rb");

	if(NULL==file)
	{
		const int err = errno;
		fprintf(stderr, "Error: Could not open '%s' for reading: %s\n", hname,
			strerror(err));
		impl->f_close(fd);
		return;
	}

	char buf[2048];
	int  size;

	while(0<(size=fread(buf, 1, sizeof(buf), file)))
	{
		if(size!=impl->f_write(fd, buf, size, &ec))
		{
			break;
		}
	}

	if(0==ec)
	{
		printf("Success\n");
	}
	else
	{
		printf("\nWrite error: %s\n", impl->f_strerror(ec));
	}

	fclose(file);
	impl->f_close(fd);
}

static void func_save(const struct impl* const impl)
{
	const char* const hname = strtok(NULL, delims);

	if(NULL==hname)
	{
		printf("Error: missing argument\n");
		return;
	}

	struct w_statfs st;

	const int ec = impl->f_statfs(&st);

	if(0!=ec)
	{
		printf("Could not stat filesystem: %s\n", impl->f_strerror(ec));
		return;
	}

	FILE* const file = fopen(hname, "wb");

	if(NULL==file)
	{
		const int err = errno;
		fprintf(stderr, "Error: Could not create '%s': %s\n", hname,
			strerror(err));
		return;
	}

	if(st.blocks==fwrite(imgtool_img_map, st.bsize, st.blocks, file))
	{
		printf("Success\n");
	}
	else
	{
		const int err = errno;
		printf("Error: Failed to write to '%s': %s\n", hname, strerror(err));
	}

	fclose(file);
}

static const struct cmd cmds[] =
{
	{
		.name = "help",
		.help = "List commands, or get help about a command",
		.func = func_help,
	},
	{
		.name = "df",
		.help = "Display filesystem statistics and info",
		.func = func_df,
	},
	{
		.name = "fsck",
		.help = "Perform filesystem check",
		.func = func_fsck,
	},
	{
		.name = "cd",
		.help = "Change directory (arg: DIR)",
		.func = func_cd,
	},
	{
		.name = "mkdir",
		.help = "Make a new directory (arg: DIR)",
		.func = func_mkdir,
	},
	{
		.name = "rmdir",
		.help = "Remove an existing directory (arg: DIR)",
		.func = func_rmdir,
	},
	{
		.name = "unlink",
		.help = "Remove an existing file (arg: WOFSFILE)",
		.func = func_unlink,
	},
	{
		.name = "ls",
		.help = "List directory content (arg: [DIR])",
		.func = func_ls,
	},
	{
		.name = "cat",
		.help = "Concatenate file to stdout (arg: WOFSFILE)",
		.func = func_cat,
	},
	{
		.name = "export",
		.help = "Export a file to the host filesystem (arg: WOFSFILE, HOSTFILE)",
		.func = func_export,
	},
	{
		.name = "import",
		.help = "Import a file from the host filesystem (arg: WOFSFILE, HOSTFILE)",
		.func = func_import,
	},
	{
		.name = "save",
		.help = "Save the modified image to the host filesystem (arg: HOSTFILE)",
		.func = func_save,
	},
	{ 0 }
};

void process_commands(const struct impl* const impl)
{
	for(;;)
	{
		printf("%s$ ", impl->f_getcwd());
		fflush(NULL);

		if(NULL==fgets(line, sizeof(line)-1, stdin))
		{
			printf("\n");
			return;
		}

		line[sizeof(line)-1] = 0;

		char* const cmdname = strtok(line, delims);

		if(NULL==cmdname || '#'==cmdname[0])
		{
			continue;
		}

		const struct cmd* const cmd = find_cmd(cmdname);

		if(NULL!=cmd)
		{
			cmd->func(impl);
		}
		else
		{
			printf("Unknown command: %s\n", cmdname);
		}
	}
}
