/*
Write-once filesystem

Copyright (c) 2022 Aron Barath
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the Author nor the names of contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Source code: svn://svn.repo.hu/wroncefs/trunk
Contact the author: aron-dev@mailbox.org
*/

#include "impl.h"
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#define wroncefs_op_write impl__op_write
#define wroncefs_op_erase impl__op_erase

#include <wroncefs.h>

#define GLUE_(a,b)      a##b
#define GLUE(a,b)       GLUE_(a,b)
#define STRINGIFY_(a)   #a
#define STRINGIFY(a)    STRINGIFY_(a)

#include "../all-src.c"

#define W_ERR_ENOMEM   1000

static int8_t impl__op_write(WOFS_STATE_ARG const void* buf,
	wofs_addr_t addr, wofs_size_t size)
{
	return imgtool_op_write(buf, addr, size, wofs_state.blockszsh);
}

static int8_t impl__op_erase(WOFS_STATE_ARG wofs_addr_t addr,
	wofs_size_t size)
{
	return imgtool_op_erase(addr, size, wofs_state.blockszsh);
}

static int wrap_off2shift(uint32_t size)
{
	return wofs_off_to_shift(size);
}

static const char* wrap_strerror(int err)
{
	if((-W_ERR_ENOMEM)==err) { return strerror(ENOMEM); }

	return wofs_strerror(-err);
}

static wofs_ino_t cwd;
static char cwdstr[8192];

static int wrap_mount(uint16_t nblocks, uint8_t blockszsh)
{
	cwd = WOFS_ROOT_INODE;
	strcpy(cwdstr, "/");
	return wofs_mount(nblocks, blockszsh, WOFS_MOUNT_ESSENTIONAL);
}

static int wrap_fsck(void)
{
	struct wofs_statfs wst;

	const int8_t ec = wofs_statfs(&wst);

	if(0!=ec)
	{
		return ec;
	}

	return wofs_fsck(wst.f_blocks, wofs_off_to_shift(wst.f_bsize), NULL);
}

static const char* wrap_getcwd(void)
{
	return cwdstr;
}

static int wrap_chdir(const char* name)
{
	if('.'==name[0] && 0==name[1])
	{
		return 0;
	}

	int up = 0;

	if('.'==name[0] && '.'==name[1] && 0==name[2])
	{
		if(WOFS_ROOT_INODE==cwd)
		{
			return 0;
		}

		up = 1;
	}

	int8_t ec;

	if(0!=(ec=wofs_chdir(&cwd, name)))
	{
		return ec;
	}

	if(up)
	{
		char* last = strrchr(cwdstr, '/');

		if(cwdstr==last)
		{
			strcpy(cwdstr, "/");
		}
		else
		if(NULL!=last)
		{
			*last = 0;
		}
	}
	else
	{
		strcat(cwdstr, name);
	}

	return 0;
}

static int wrap_mkdir(const char* name)
{
	return wofs_mkdir(cwd, name, NULL);
}

static int wrap_rmdir(const char* name)
{
	return wofs_rmdir(cwd, name);
}

static int wrap_unlink(const char* name)
{
	return wofs_unlink(cwd, name);
}

static int wrap_statfs(struct w_statfs* const st)
{
	struct wofs_statfs wst;

	const int8_t ec = wofs_statfs(&wst);

	st->bsize   = wst.f_bsize;
	st->bfree   = wst.f_bfree;
	st->blocks  = wst.f_blocks;
	st->namelen = wst.f_namelen;

	return 0;
}

struct w_dir
{
	struct wofs_dir dir;
	struct w_dirent ent;
	wofs_ino_t      ino;
};

static struct w_dir* wrap_opendir(const char* name, int* const out_ec)
{
	struct w_dir* const dir = (struct w_dir*)malloc(sizeof(struct w_dir));

	if(NULL==dir)
	{
		*out_ec = -W_ERR_ENOMEM;
		return NULL;
	}

	dir->ino = cwd;

	if(0!=((*out_ec)=wofs_chdir(&dir->ino, name)))
	{
		free(dir);
		return NULL;
	}

	if(0!=((*out_ec)=wofs_opendir(&dir->dir, dir->ino)))
	{
		free(dir);
		return NULL;
	}

	return dir;
}

static struct w_dirent* wrap_readdir(struct w_dir* const dir,
	int* const out_ec)
{
	int8_t ec;
	struct wofs_dirent* const ent = wofs_readdir(&dir->dir, &ec);

	if((-WOFS_ERR_ENOMORE)==ec)
	{
		*out_ec = 0;
	}
	else
	{
		*out_ec = ec;
	}

	if(NULL==ent)
	{
		return NULL;
	}

	dir->ent.ino     = ent->d_ino;
	dir->ent.namelen = ent->d_namelen;
	dir->ent.name    = ent->d_name;

	return &dir->ent;
}

static void wrap_closedir(struct w_dir* const dir)
{
	free(dir);
}

static int wrap_istat(size_t ino, struct w_stat* const st)
{
	int8_t  ec;
	uint8_t type;

	if(0!=(ec=wofs_istat(ino, &type)))
	{
		return ec;
	}

	if(WOFS_INOTYPE_DIR==type)
	{
		st->is_file = 0;
		st->size    = 0;
	}
	else
	if(WOFS_INOTYPE_FILE==type)
	{
		struct wofs_fd fd;

		if(0!=(ec=wofs_iopen(&fd, ino)))
		{
			return ec;
		}

		wofs_off_t off = wofs_seek(&fd, 0, WOFS_SEEK_END, &ec);

		if(0>off)
		{
			return ec;
		}

		st->is_file = 1;
		st->size    = off;
	}
	else
	{
		return -WOFS_ERR_EBADINODE;
	}

	return 0;
}

struct w_file
{
	struct wofs_fd fd;
};

static struct w_file* wrap_open(const char* name, int* const out_ec)
{
	struct w_file* const fd = (struct w_file*)malloc(sizeof(struct w_file));

	if(NULL==fd)
	{
		*out_ec = -W_ERR_ENOMEM;
		return NULL;
	}

	if(0!=((*out_ec)=wofs_open(&fd->fd, cwd, name)))
	{
		free(fd);
		return NULL;
	}

	return fd;
}

static struct w_file* wrap_creat(const char* name, int* const out_ec)
{
	struct w_file* const fd = (struct w_file*)malloc(sizeof(struct w_file));

	if(NULL==fd)
	{
		*out_ec = -W_ERR_ENOMEM;
		return NULL;
	}

	if(0!=((*out_ec)=wofs_creat(&fd->fd, cwd, name)))
	{
		free(fd);
		return NULL;
	}

	return fd;
}

static int wrap_read(struct w_file* const fd, void* buf, int size,
	int* const out_ec)
{
	int8_t ec;
	const int res = wofs_read(&fd->fd, buf, size, &ec);

	if(NULL!=out_ec)
	{
		*out_ec = ec;
	}

	return res;
}

static int wrap_write(struct w_file* const fd, const void* buf, int size,
	int* const out_ec)
{
	int8_t ec;
	const int res = wofs_write(&fd->fd, buf, size, &ec);

	if(NULL!=out_ec)
	{
		*out_ec = ec;
	}

	return res;
}

static int wrap_seek(struct w_file* const fd, int offs, int whence)
{
	switch(whence)
	{
	case W_SEEK_SET: whence = WOFS_SEEK_SET; break;
	case W_SEEK_CUR: whence = WOFS_SEEK_CUR; break;
	case W_SEEK_END: whence = WOFS_SEEK_END; break;
	default: return -WOFS_ERR_EINVAL;
	}

	return wofs_seek(&fd->fd, offs, whence, NULL);
}

static void wrap_close(struct w_file* const fd)
{
	free(fd);
}

const struct impl GLUE(impl_, IMPL_NAME) =
{
	.impl_name      = STRINGIFY(IMPL_NAME),
	.max_filesize   = WOFS_MAX_FILESIZE,
	.min_blockszsh  = WOFS_MIN_BLOCKSZSH,
	.max_blockszsh  = WOFS_MAX_BLOCKSZSH,
	.inf_can_alter  = WRONCEFS_CAN_ALTER,
	.inf_16bit_mode = WRONCEFS_16BIT_MODE,

	.f_off2shift = wrap_off2shift,
	.f_strerror  = wrap_strerror,

	.f_mount     = wrap_mount,
	.f_fsck      = wrap_fsck,
	.f_getcwd    = wrap_getcwd,
	.f_chdir     = wrap_chdir,
	.f_mkdir     = wrap_mkdir,
	.f_rmdir     = wrap_rmdir,
	.f_unlink    = wrap_unlink,
	.f_statfs    = wrap_statfs,
	.f_istat     = wrap_istat,

	.f_opendir   = wrap_opendir,
	.f_readdir   = wrap_readdir,
	.f_closedir  = wrap_closedir,

	.f_open      = wrap_open,
	.f_creat     = wrap_creat,
	.f_read      = wrap_read,
	.f_write     = wrap_write,
	.f_seek      = wrap_seek,
	.f_close     = wrap_close,
};
