/*
Write-once filesystem

Copyright (c) 2022 Aron Barath
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the Author nor the names of contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Source code: svn://svn.repo.hu/wroncefs/trunk
Contact the author: aron-dev@mailbox.org
*/

#include "impl.h"
#include "cmd.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

void* imgtool_img_map;
static size_t img_size;

int8_t imgtool_op_write(const void* buf, uint32_t addr, uint32_t size,
	uint8_t blockszsh)
{
	if((addr+size)>img_size)
	{
		return -1;
	}

	memcpy(((uint8_t*)imgtool_img_map)+addr, buf, size);

	return 0;
}

int8_t imgtool_op_erase(uint32_t addr, uint32_t size, uint8_t blockszsh)
{
	if((addr+size)>img_size)
	{
		return -1;
	}

	memset(((uint8_t*)imgtool_img_map)+addr, -1, size);

	return 0;
}

const struct impl* impls[] =
{
	&impl_alter16,
	&impl_alter32,
	&impl_noalter16,
	&impl_noalter32,
	0
};

static void remap(const int fd)
{
	if(NULL!=imgtool_img_map)
	{
		munmap(imgtool_img_map, img_size);
	}

	imgtool_img_map = mmap(NULL, img_size, PROT_READ|PROT_WRITE,
		MAP_PRIVATE, fd, 0);

	if(MAP_FAILED==imgtool_img_map)
	{
		const int err = errno;
		fprintf(stderr, "Error: Could not (re)map image: %s\n",
			strerror(err));
		exit(1);
	}
}

int main(int argc, char* argv[])
{
	const char* const progname = argv[0];

	++argv;
	--argc;

	int hint_alter   = -1;
	int hint_16bit   = -1;
	int hint_blkszsh = -1;

	while(2<=argc) /* every option has an argument */
	{
		if(0==strcmp(argv[0], "-a"))
		{
			if(0==strcmp(argv[1], "yes"))
			{
				hint_alter = 1;
			}
			else
			if(0==strcmp(argv[1], "no"))
			{
				hint_alter = 0;
			}
			else
			{
				goto bad_arg;
			}
		}
		else
		if(0==strcmp(argv[0], "-i"))
		{
			if(0==strcmp(argv[1], "16"))
			{
				hint_16bit = 1;
			}
			else
			if(0==strcmp(argv[1], "32"))
			{
				hint_16bit = 0;
			}
			else
			{
				goto bad_arg;
			}
		}
		else
		if(0==strcmp(argv[0], "-b"))
		{
			hint_blkszsh = atoi(argv[1]);
			if(0>=hint_blkszsh) { goto bad_arg; }
		}
		else
		if(0==strcmp(argv[0], "-B"))
		{
			hint_blkszsh = impl_alter32.f_off2shift(atoi(argv[1]));
			if(0>=hint_blkszsh) { goto bad_arg; }
		}
		else
		{
			fprintf(stderr, "%s: Unknown option: '%s'\n", progname, argv[0]);
			exit(1);

		bad_arg:;
			fprintf(stderr, "%s: Bad argument to '%s'\n", progname, argv[0]);
			exit(1);
		}

		argv += 2;
		argc -= 2;
	}

	if(1!=argc)
	{
		fprintf(stderr, "Usage: %s [OPTION...] FILE\n", progname);
		static const char help_text[] =
		"\n"
		"OPTIONs:\n"
		"  -a 'yes'/'no'  try mount only in 'alter'/'non-alter' mode\n"
		"  -i '16'/'32'   try mount only in '16-bit'/'32-bit' mode\n"
		"  -b BITS        set block size to \"1 << BITS\"\n"
		"  -B SIZE        set block size to SIZE (must be power of 2)\n"
		;
		fputs(help_text, stderr);
		exit(1);
	}

	const char* const imgfname = argv[0];

	const int fd = open(argv[0], O_RDONLY|O_CLOEXEC|O_NOCTTY);

	if(0>fd)
	{
		const int err = errno;
		fprintf(stderr, "Error: Could not open '%s' for read: %s\n", imgfname,
			strerror(err));
		exit(1);
	}

	struct stat st;

	if(0!=fstat(fd, &st))
	{
		const int err = errno;
		fprintf(stderr, "Error: Could not stat fd of '%s': %s\n", imgfname,
			strerror(err));
		exit(1);
	}

	if(SIZE_MAX<st.st_size)
	{
		fprintf(stderr, "Error: File '%s' is too large\n", imgfname);
		exit(1);
	}

	img_size = (size_t)st.st_size;

	for(const struct impl** p_impl=impls;*p_impl;++p_impl)
	{
		const struct impl* const impl = *p_impl;

		/* 'alter' mode was requested, but implementation is 'non-alter' */
		if(1==hint_alter && !impl->inf_can_alter)
		{
			continue;
		}

		/* 'non-alter' mode was requested, but implementation is 'alter' */
		if(0==hint_alter && impl->inf_can_alter)
		{
			continue;
		}

		/* 16-bit mode was requested, but implementation is 32-bit */
		if(1==hint_16bit && !impl->inf_16bit_mode)
		{
			continue;
		}

		/* 32-bit mode was requested, but implementation is 16-bit */
		if(0==hint_16bit && impl->inf_16bit_mode)
		{
			continue;
		}

		if(0>hint_blkszsh)
		{
			for(uint32_t szsh=impl->min_blockszsh;szsh<impl->max_blockszsh;
				++szsh)
			{
				remap(fd);

				if(0==impl->f_mount(img_size >> szsh, szsh))
				{
					process_commands(impl);
					return 0;
				}
			}
		}
		else
		{
			remap(fd);

			if(0==impl->f_mount(img_size >> hint_blkszsh, hint_blkszsh))
			{
				process_commands(impl);
				return 0;
			}
		}
	}

	fprintf(stderr, "Error: Could not mount '%s'\n", imgfname);
	return 1;
}
