/*
Write-once filesystem

Copyright (c) 2022 Aron Barath
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the Author nor the names of contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Source code: svn://svn.repo.hu/wroncefs/trunk
Contact the author: aron-dev@mailbox.org
*/

#include "impl.h"
#include <string.h>
#include <stddef.h>

static int8_t fix_length_indicator(WOFS_STATE_ARG struct wofs_fd* const fd,
	uint16_t plim)
{
	const wofs_addr_t start = fd->fd_addr + sizeof(struct wofs_inode) +
		WOFS_SIZEOF_FILEINFO;

	/* end of the possible data area -- also the start address of the */
	/* length indicator */
	const wofs_addr_t end = fd->fd_addr + plim;

	wofs_addr_t addr = start;

	/* address of the last non-0xFF byte -- if this is not set inside the */
	/* loop above, we must assume that the block is completely full */
	wofs_addr_t last = end;

	while(addr<end)
	{
		WOFS_READ_LOCAL(uint8_t, data, addr);

		if(0xFF!=WOFS_LOCAL_VALUE(data))
		{
			last = addr;
		}

		++addr;
	}

	const uint16_t clen = last - start;
	fd->fd_clen = clen;

#if WRONCEFS_CAN_ALTER

	/* write slot #2 first */
	if(0!=wroncefs_op_write(PASS_STATE &clen, end+sizeof(clen), sizeof(clen)))
	{
		return -WOFS_ERR_EIOW;
	}

	/* write slot #1 second */
	if(0!=wroncefs_op_write(PASS_STATE &clen, end, sizeof(clen)))
	{
		return -WOFS_ERR_EIOW;
	}

#else /* WRONCEFS_CAN_ALTER */

	if((clen+sizeof(uint16_t)+sizeof(uint8_t))<=(end-start))
	{
		if(0!=wroncefs_op_write(PASS_STATE &clen, end, sizeof(clen)))
		{
			return -WOFS_ERR_EIOW;
		}

		const uint8_t aux = WOFS_LENGTHIND_VALID;

		if(0!=wroncefs_op_write(PASS_STATE &aux, end-sizeof(aux),
			sizeof(aux)))
		{
			return -WOFS_ERR_EIOW;
		}

		fd->fd_plim = plim - (sizeof(uint16_t)+sizeof(uint8_t));
	}
	else
	{
		return wofs__file_rewrite(PASS_STATE fd, clen);
	}

#endif /* WRONCEFS_CAN_ALTER */

	return 0;
}

WOFS_IMPL int8_t wofs__fd_setup(WOFS_STATE_ARG struct wofs_fd* const fd,
	const wofs_ino_t ino)
{
	const uint8_t shift = WOFS_STATE.blockszsh;

	fd->fd_addr = ((wofs_addr_t)WOFS_INODE_TO_BLOCK(ino)) << shift;
	fd->fd_lpos = 0;

	{
		WOFS_READ_LOCAL(wofs_off_t, start_offs,
			fd->fd_addr+sizeof(struct wofs_inode)+
			offsetof(struct wofs_fileinfo, start_offs));

		fd->fd_fpos = WOFS_LOCAL_VALUE(start_offs);
	}

#if WRONCEFS_CAN_ALTER

	{
		const uint16_t plim =
			(uint16_t)((((wofs_addr_t)1) << shift)-
				sizeof(struct wofs_lengthind));

		WOFS_READ_LOCAL(struct wofs_lengthind, li, fd->fd_addr+plim);

		if(WOFS_LOCAL_FIELD(li, len1)!=WOFS_LOCAL_FIELD(li, len2))
		{
			const int8_t ec = fix_length_indicator(PASS_STATE fd, plim);

			if(0!=ec)
			{
				return ec;
			}
		}
		else
		{
			fd->fd_clen = WOFS_LOCAL_FIELD(li, len2);
		}

		if(0xFFFF==fd->fd_clen)
		{
			fd->fd_clen = 0;
		}
	}

#else /* WRONCEFS_CAN_ALTER */

	uint16_t plim = (uint16_t)((((wofs_addr_t)1) << shift)-sizeof(uint16_t));
	uint16_t lisz = sizeof(uint16_t);

	fd->fd_clen = 0;

	while((sizeof(struct wofs_inode)+WOFS_SIZEOF_FILEINFO)<plim)
	{
		struct wofs_lengthind li;

		/* 'plim' is possibly unaligned */
		WOFS_READ_EXTERN(&li, fd->fd_addr+plim, lisz);

		if(0xFFFF==li.length)
		{
			if(sizeof(uint16_t)<lisz && WOFS_LENGTHIND_VALID!=li.valid)
			{
				const int8_t ec = fix_length_indicator(PASS_STATE fd, plim);

				if(0!=ec)
				{
					return ec;
				}

				plim = fd->fd_plim;
			}

			break;
		}

		lisz = sizeof(uint16_t) + sizeof(uint8_t);

		fd->fd_clen = li.length;
		plim -= lisz;
	}

	if((sizeof(struct wofs_inode)+WOFS_SIZEOF_FILEINFO)>=plim)
	{
		return -WOFS_ERR_ENOTRECOVERABLE;
	}

	fd->fd_plim = plim;

#endif /* WRONCEFS_CAN_ALTER */

	return 0;
}
