/*
Write-once filesystem

Copyright (c) 2022 Aron Barath
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the Author nor the names of contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Source code: svn://svn.repo.hu/wroncefs/trunk
Contact the author: aron-dev@mailbox.org
*/

#include "impl.h"
#include <stddef.h>

static int8_t find_empty_block(WOFS_STATE_ARG uint16_t* const out_blk,
	wofs_addr_t* const out_addr)
{
	const uint16_t nblocks = WOFS_STATE.nblocks;
	const uint8_t  shift   = WOFS_STATE.blockszsh;

	/* we could start from block 0, but it is reserved for the root */
	/* directory, thus we don't need to check it */
	for(uint16_t blk=1;blk<nblocks;++blk)
	{
		const wofs_addr_t addr = ((wofs_addr_t)blk) << shift;

		WOFS_READ_LOCAL(struct wofs_inode, inode, addr);

		/* if the loaded inode structure contains only 0xFF bytes, we */
		/* accept it as empty */

		uint8_t*       ptr = (uint8_t*)WOFS_LOCAL_ADDR(inode);
		uint8_t* const end = ptr + sizeof(struct wofs_inode);

		while(ptr<end && 0xFF==(*ptr)) { ++ptr; }

		if(ptr==end)
		{
			*out_blk  = blk;
			*out_addr = addr;
			return 0;
		}
	}

	return -WOFS_ERR_ENOSPC;
}

WOFS_IMPL int8_t wofs__alloc_inode(WOFS_STATE_ARG wofs_ino_t* const out_ino)
{
	uint16_t    blk;
	wofs_addr_t addr;

	if(0==WOFS_STATE.freeblocks)
	{
		return -WOFS_ERR_ENOSPC;
	}

	const int8_t ec = find_empty_block(PASS_STATE &blk, &addr);

	if(0!=ec)
	{
		return ec;
	}

	--WOFS_STATE.freeblocks;
	*out_ino = WOFS_BLOCK_TO_INODE(blk);

	return 0;
}

WOFS_IMPL int8_t wofs__get_temp_block(WOFS_STATE_ARG wofs_addr_t* const addr)
{
	uint16_t blk_id;

	return find_empty_block(PASS_STATE &blk_id, addr);
}
