/*
Write-once filesystem

Copyright (c) 2022 Aron Barath
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the Author nor the names of contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Source code: svn://svn.repo.hu/wroncefs/trunk
Contact the author: aron-dev@mailbox.org
*/

#include "impl.h"
#include <string.h>
#include <stddef.h>

static int8_t restore_fsinfo(WOFS_STATE_ARG const uint16_t nblocks,
	const uint8_t blockszsh)
{
	int8_t ec;

	if(0!=(ec=wofs__check_inode(PASS_STATE WOFS_ROOT_INODE,
		WOFS_INOTYPE_DIR)))
	{
		return ec;
	}

	{
		WOFS_READ_LOCAL(struct wofs_rawdirent, de, sizeof(struct wofs_inode));

		if(WOFS_INVALID_INODE!=WOFS_LOCAL_FIELD(de, d_ino) ||
			sizeof(struct wofs_fsinfo)!=WOFS_LOCAL_FIELD(de, d_namelen) ||
			WOFS_DIRENT_FSINFO!=WOFS_LOCAL_FIELD(de, d_alive))
		{
			return -WOFS_ERR_EBADINFO;
		}
	}

	{
		WOFS_READ_LOCAL(struct wofs_fsinfo, fi,
			sizeof(struct wofs_inode)+(sizeof(struct wofs_rawdirent)-
			sizeof(((struct wofs_rawdirent*)0)->d_name2)));

		if(blockszsh!=WOFS_LOCAL_FIELD(fi, fsi_blockszsh) ||
			WRONCEFS_NAME_MAX<WOFS_LOCAL_FIELD(fi, fsi_maxnamelen) ||
			nblocks!=WOFS_LOCAL_FIELD(fi, fsi_nblocks) ||
			wofs_fsi_flags!=WOFS_LOCAL_FIELD(fi, fsi_flags) ||
			0xFFFF!=WOFS_LOCAL_FIELD(fi, fsi_reserved))
		{
			return -WOFS_ERR_EBADINFO;
		}

		WOFS_STATE.name_max = WOFS_LOCAL_FIELD(fi, fsi_maxnamelen);
	}

	return 0;
}

static int8_t fix_dirent_rem(WOFS_STATE_ARG const uint16_t nblocks,
	const uint8_t blockszsh)
{
	int8_t ec;

	const wofs_addr_t tdi_offs =
		(((wofs_addr_t)1) << blockszsh) - sizeof(struct wofs_tempdirinfo);

	const uint16_t max_content_size =
		(((wofs_size_t)1) << blockszsh) - sizeof(struct wofs_inode) -
		sizeof(struct wofs_tempdirinfo);

	for(wofs_ino_t ino=1;ino<=nblocks;++ino)
	{
		if(0==wofs__check_inode(PASS_STATE ino, WOFS_INOTYPE_TMPDIR))
		{
			const wofs_addr_t temp_addr =
				((wofs_addr_t)WOFS_INODE_TO_BLOCK(ino)) << blockszsh;

			WOFS_READ_LOCAL(struct wofs_tempdirinfo, tdi, temp_addr+tdi_offs);

			if(max_content_size<WOFS_LOCAL_FIELD(tdi, cont_size) ||
				0==WOFS_LOCAL_FIELD(tdi, orig_ino) ||
				nblocks<WOFS_LOCAL_FIELD(tdi, orig_ino) ||
				ino==WOFS_LOCAL_FIELD(tdi, orig_ino))
			{
				/* "content size" or "original inode" is garbage, retreat */
				/* before messing things up */
				return -WOFS_ERR_ENOTRECOVERABLE;
			}

			const wofs_addr_t orig_addr =
				((wofs_addr_t)WOFS_INODE_TO_BLOCK(WOFS_LOCAL_FIELD(tdi,
				orig_ino))) << blockszsh;

			struct wofs_inode inode;
			WOFS_READ_EXTERN(&inode, temp_addr, sizeof(inode));

			/* since this inode is a temporary copy of a *directory* */
			/* block, restore 'i_type' */
			inode.i_type = WOFS_INOTYPE_DIR;

			if(0!=wroncefs_op_erase(PASS_STATE orig_addr,
				((wofs_size_t)1) << blockszsh))
			{
				return -WOFS_ERR_EIOE;
			}

			if(0!=(ec=wofs__ino_fill_1(PASS_STATE &inode, orig_addr)))
			{
				return ec;
			}

			if(0!=(ec=wofs__copy_storage_content(PASS_STATE
				orig_addr+sizeof(struct wofs_inode),
				temp_addr+sizeof(struct wofs_inode),
				WOFS_LOCAL_FIELD(tdi, cont_size))))
			{
				return ec;
			}

			if(0!=(ec=wofs__ino_fill_2(PASS_STATE &inode, orig_addr)))
			{
				return ec;
			}

			if(0!=wroncefs_op_erase(PASS_STATE temp_addr,
				((wofs_size_t)1) << blockszsh))
			{
				return -WOFS_ERR_EIOE;
			}
		}
	}

	return 0;
}

static int8_t restore_temporary_inodes(WOFS_STATE_ARG const uint16_t nblocks,
	const uint8_t blockszsh, const uint8_t do_restore)
{
	int8_t ec;

	for(wofs_ino_t ino=1;ino<=nblocks;++ino)
	{
		if(0!=wofs__check_inode(PASS_STATE ino, 0))
		{
			continue;
		}

		const wofs_addr_t addr =
			((wofs_addr_t)WOFS_INODE_TO_BLOCK(ino)) << blockszsh;

		WOFS_READ_LOCAL(uint8_t, i_type,
			addr+offsetof(struct wofs_inode, i_type));

		switch(WOFS_LOCAL_VALUE(i_type))
		{
		case WOFS_INOTYPE_TMPFILE:
			if(!do_restore)
			{
				/* if not temporary file restoration was requested, we */
				/* must fail, because the filesystem is in an unstable */
				/* state */
				return -WOFS_ERR_EUCLEAN;
			}
			if(0!=(ec=wofs__file_restore_temp(PASS_STATE ino)))
			{
				return ec;
			}
			break;

		case WOFS_INOTYPE_TMPDIR:
			/* check whether TEMPDIR was processed earlier -- if not */
			/* (because it was not requested to do), we must fail */
			return -WOFS_ERR_EUCLEAN;

		default:
			break;
		};
	}

	return 0;
}

static int8_t count_free_inodes(WOFS_STATE_ARG const uint16_t nblocks,
	const uint8_t blockszsh, const uint8_t erase_garbage)
{
	int8_t ec;

	WOFS_STATE.freeblocks = 0;

	for(wofs_ino_t ino=1;ino<=nblocks;++ino)
	{
		const wofs_addr_t addr =
			((wofs_addr_t)WOFS_INODE_TO_BLOCK(ino)) << blockszsh;

		if(0==wofs__check_inode(PASS_STATE ino, 0))
		{
			/* valid inode, which is not free */
			continue;
		}

		WOFS_READ_LOCAL(struct wofs_inode, inode, addr);

		/* it is adequate to check only whether the whole inode structure */
		/* is 0xFF */

		uint8_t*       ptr = (uint8_t*)WOFS_LOCAL_ADDR(inode);
		uint8_t* const end = ptr + sizeof(struct wofs_inode);

		while(ptr<end && 0xFF==(*ptr)) { ++ptr; }

		if(ptr==end)
		{
			/* yes, it's full of 0xFF */
			++WOFS_STATE.freeblocks;
		}
		else
		if(erase_garbage)
		{
			if(0!=wroncefs_op_erase(PASS_STATE addr,
				((wofs_size_t)1)<<blockszsh))
			{
				return -WOFS_ERR_EIOE;
			}

			/* now it's full of 0xFF */
			++WOFS_STATE.freeblocks;
		}
	}

	if(WOFS_NUM_SPARE_BLOCKS>WOFS_STATE.freeblocks)
	{
		return -WOFS_ERR_ENOTRECOVERABLE;
	}

	WOFS_STATE.freeblocks -= WOFS_NUM_SPARE_BLOCKS;

	return 0;
}

WOFS_IMPL int8_t wofs__mount_pre(WOFS_STATE_ARG uint16_t nblocks,
	uint8_t blockszsh, uint16_t flags)
{
	int8_t  ec;
	uint8_t fsinfo = 0;

	memset(&WOFS_STATE, 0, sizeof(WOFS_STATE));

	if(WOFS_MIN_NBLOCKS>nblocks || WOFS_MAX_NBLOCKS<nblocks)
	{
		return -WOFS_ERR_EINVAL;
	}

	if(WOFS_MIN_BLOCKSZSH>blockszsh || WOFS_MAX_BLOCKSZSH<blockszsh)
	{
		return -WOFS_ERR_EINVAL;
	}

	/* we need to set these, because following operations rely on them */

	WOFS_STATE.nblocks   = nblocks;
	WOFS_STATE.blockszsh = blockszsh;

	if(0==wofs__check_inode(PASS_STATE WOFS_ROOT_INODE, 0))
	{
		/* root inode might be moved into a temporary node */

		if(0!=(ec=restore_fsinfo(PASS_STATE nblocks, blockszsh)))
		{
			return ec;
		}

		fsinfo = 1;
	}

	if(WOFS_MOUNT_FIXDIRENTREM & flags)
	{
		/* fix interrupted directory entry removal -- note: 'nblocks' and */
		/* 'blockszsh' might not be validated here, but we must rely on */
		/* them in order to locate the (possibly) moved root directory */

		if(0!=(ec=fix_dirent_rem(PASS_STATE nblocks, blockszsh)))
		{
			return ec;
		}

		/* restore fsinfo if it wasn't restored earlier -- note: we don't */
		/* call check_inode() with soft-error return */

		if(!fsinfo && 0!=(ec=restore_fsinfo(PASS_STATE nblocks, blockszsh)))
		{
			return ec;
		}
	}
	else
	if(!fsinfo)
	{
		/* no fix was requested, and we can't mount it */
		return -WOFS_ERR_ENOTRECOVERABLE;
	}

	/* try to restore temporary inodes, but fail, if restoration was not */
	/* requested */

	if(0!=(ec=restore_temporary_inodes(PASS_STATE nblocks, blockszsh,
		0!=(WOFS_MOUNT_RESTRTEMPFILE & flags))))
	{
		return ec;
	}

	return 0;
}

WOFS_IMPL int8_t wofs__mount_post(WOFS_STATE_ARG uint16_t nblocks,
	uint8_t blockszsh, uint16_t flags)
{
	int8_t ec;

	if(0!=(ec=count_free_inodes(PASS_STATE nblocks, blockszsh,
		0!=(WOFS_MOUNT_ERASEGARBAGE & flags))))
	{
		return ec;
	}

	return 0;
}

WOFS_IMPL int8_t wofs_mount(WOFS_STATE_ARG uint16_t nblocks,
	uint8_t blockszsh, uint16_t flags)
{
	int8_t ec;

	if(0!=(ec=wofs__mount_pre(PASS_STATE nblocks, blockszsh, flags)) ||
		0!=(ec=wofs__mount_post(PASS_STATE nblocks, blockszsh, flags)))
	{
		memset(&WOFS_STATE, 0, sizeof(WOFS_STATE));
		return ec;
	}

	return 0;
}
