/*
Write-once filesystem

Copyright (c) 2022 Aron Barath
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the Author nor the names of contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Source code: svn://svn.repo.hu/wroncefs/trunk
Contact the author: aron-dev@mailbox.org
*/

#include "impl.h"
#include <stddef.h>
#include <string.h>

#define FILE_INODE_ADMINISTRATION_SIZE \
	(sizeof(struct wofs_inode) + WOFS_SIZEOF_FILEINFO)

static wofs_size_t compute_write_part_size(WOFS_STATE_ARG
	struct wofs_fd* const fd, wofs_size_t size, const uint8_t shift)
{
#if WRONCEFS_CAN_ALTER

	size_t space = (((wofs_size_t)1)<<shift) - WOFS_SIZEOF_LENGTHIND -
		FILE_INODE_ADMINISTRATION_SIZE - fd->fd_lpos;

#else /* WRONCEFS_CAN_ALTER */

	(void)shift;

	if(0<fd->fd_clen && (fd->fd_clen-fd->fd_lpos)>=size)
	{
		/* current write fits into the preallocated space */
		return size;
	}

	/* we need to ensure that the last WOFS_SIZEOF_LENGTHIND octets are */
	/* preserved for the file inode fixup -- but we will already preserved */
	/* an uint16_t there */

	size_t space = (((wofs_size_t)1)<<shift) -
		(WOFS_SIZEOF_LENGTHIND - sizeof(uint16_t));

	if(space>fd->fd_plim)
	{
		space = fd->fd_plim;
	}

	space -= (FILE_INODE_ADMINISTRATION_SIZE + fd->fd_lpos);

	if((sizeof(uint16_t)+sizeof(uint8_t))>=space)
	{
		/* we can't allocate a new length indicator */
		return 0;
	}

	/* we need to allocate a new length indicator */

	space -= (sizeof(uint16_t) + sizeof(uint8_t));

#endif /* WRONCEFS_CAN_ALTER */

	if(space>size)
	{
		space = size;
	}

	return space;
}

static int8_t update_length_indicator_by(WOFS_STATE_ARG
	struct wofs_fd* const fd, wofs_size_t addend, const uint8_t shift)
{
	const uint16_t new_len = fd->fd_lpos + addend;

	if(new_len<=fd->fd_clen)
	{
		/* no update is needed */
		return 0;
	}

	const wofs_addr_t li_addr =
#if WRONCEFS_CAN_ALTER
		(((wofs_size_t)1)<<shift) - sizeof(struct wofs_lengthind)
#else /* WRONCEFS_CAN_ALTER */
		fd->fd_plim
#endif /* WRONCEFS_CAN_ALTER */
		;

#if WRONCEFS_CAN_ALTER
	/* write slot #2 first */
	if(0!=wroncefs_op_write(PASS_STATE &new_len,
		fd->fd_addr+li_addr+sizeof(new_len), sizeof(new_len)))
	{
		return -WOFS_ERR_EIOW;
	}
#endif /* WRONCEFS_CAN_ALTER */

	if(0!=wroncefs_op_write(PASS_STATE &new_len, fd->fd_addr+li_addr,
		sizeof(new_len)))
	{
		return -WOFS_ERR_EIOW;
	}

#if !WRONCEFS_CAN_ALTER
	const uint8_t aux = WOFS_LENGTHIND_VALID;

	if(0!=wroncefs_op_write(PASS_STATE &aux, fd->fd_addr+li_addr-sizeof(aux),
		sizeof(aux)))
	{
		return -WOFS_ERR_EIOW;
	}
#endif /* !WRONCEFS_CAN_ALTER */

	fd->fd_clen = new_len;
#if !WRONCEFS_CAN_ALTER
	fd->fd_plim -= (sizeof(new_len) + sizeof(aux));
#endif /* !WRONCEFS_CAN_ALTER */

	return 0;
}

static int8_t step_or_extend_file(WOFS_STATE_ARG struct wofs_fd* const fd,
	const uint8_t shift)
{
	int8_t ec;

	WOFS_READ_LOCAL(wofs_ino_t, first,
		fd->fd_addr+offsetof(struct wofs_inode, i_first));

	ec = wofs__follow_file_chain__forward(PASS_STATE fd);

	if((-WOFS_ERR_ENOMORE)==ec)
	{
		/* no next block, allocate one */

		struct wofs_inode inode;
		wofs_ino_t        next;

		inode.i_sig   = WOFS_SIG;
		inode.i_first = WOFS_LOCAL_VALUE(first);
		inode.i_next  = WOFS_END_INODE;
		inode.i_vnext = 0xFF;
		inode.i_resd  = 0xFF;
		inode.i_type  = WOFS_INOTYPE_FILE;
		inode.i_alive = 0xFF;

		if(0!=(ec=wofs__alloc_inode(PASS_STATE &next)))
		{
			return ec;
		}

		const wofs_addr_t addr =
			((wofs_addr_t)WOFS_INODE_TO_BLOCK(next)) << shift;

		if(0!=(ec=wofs__ino_fill_1(PASS_STATE &inode, addr)))
		{
			return ec;
		}

		if(0!=(ec=wofs__fileinfo_write(PASS_STATE addr, fd->fd_fpos,
			(wofs_ino_t)WOFS_BLOCK_TO_INODE(fd->fd_addr >> shift))))
		{
		erase_and_exit:;
			wroncefs_op_erase(PASS_STATE addr, ((wofs_size_t)1) << shift);
			--WOFS_STATE.freeblocks;
			return ec;
		}

		if(0!=wroncefs_op_write(PASS_STATE &next,
			fd->fd_addr+offsetof(struct wofs_inode, i_next), sizeof(next)))
		{
			ec = -WOFS_ERR_EIOW;
			goto erase_and_exit;
		}

		if(0!=(ec=wofs__ino_fill_2(PASS_STATE &inode, addr)))
		{
			goto erase_and_exit;
		}

		uint8_t vnext = 0;

		if(0!=wroncefs_op_write(PASS_STATE &vnext,
			fd->fd_addr+offsetof(struct wofs_inode, i_vnext), sizeof(vnext)))
		{
			ec = -WOFS_ERR_EIOW;
			goto erase_and_exit;
		}

		return wofs__fd_setup(PASS_STATE fd, next);
	}

	return ec;
}

WOFS_IMPL wofs_ssize_t wofs_write(WOFS_STATE_ARG struct wofs_fd* const fd,
	const void* buf, wofs_size_t size, int8_t* const out_ec)
{
	return wofs__write_impl(PASS_STATE fd, buf, size, out_ec, 1);
}

WOFS_IMPL wofs_ssize_t wofs__write_impl(WOFS_STATE_ARG
	struct wofs_fd* const fd, const void* buf, wofs_size_t size,
	int8_t* const out_ec, const uint8_t write_it)
{
	wofs_ssize_t res = 0;
	int8_t       ec  = 0;

	const uint8_t shift = WOFS_STATE.blockszsh;

	// TODO: sync inode?

	while(0<size)
	{
		wofs_size_t part = WOFS_MAX_FILESIZE - fd->fd_fpos;

		if(part>size)
		{
			part = size;
		}
		else
		if(0==part)
		{
			ec = -WOFS_ERR_EFBIG;
			break;
		}

		part = compute_write_part_size(PASS_STATE fd, part, shift);

		if(0==part)
		{
			if(0!=(ec=step_or_extend_file(PASS_STATE fd, shift)))
			{
				break;
			}

			continue;
		}

		if(0!=(ec=update_length_indicator_by(PASS_STATE fd, part, shift)))
		{
			break;
		}

		if(write_it && 0!=wroncefs_op_write(PASS_STATE buf,
			fd->fd_addr+fd->fd_lpos+FILE_INODE_ADMINISTRATION_SIZE, part))
		{
			ec = -WOFS_ERR_EIOW;
			break;
		}

		fd->fd_fpos += part;
		fd->fd_lpos += part;
		res += part;
		buf = ((const uint8_t*)buf) + part;
		size -= part;
	}

	if(NULL!=out_ec)
	{
		*out_ec = ec;
	}

	return res;
}
