/*
Write-once filesystem

Copyright (c) 2022 Aron Barath
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the Author nor the names of contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Source code: svn://svn.repo.hu/wroncefs/trunk
Contact the author: aron-dev@mailbox.org
*/

#include "wroncefs.h"
#include "run.h"
#include "io.h"
#include <stdio.h>
#include <string.h>

static const char extremly_long_name[] =
	"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
	"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
	"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
	"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
	;

int run_tests(const uint16_t nblocks, const uint8_t blkszsh)
{
	const wofs_size_t block_size = ((wofs_size_t)1) << blkszsh;

	wofs_ino_t wd = WOFS_ROOT_INODE;

	uint8_t all_bytes[256];

	for(size_t i=0;i<sizeof(all_bytes);++i)
	{
		all_bytes[i] = i;
	}

	CHECK(0, wofs_format(nblocks, blkszsh, 0));

	/* in root both "." and ".." points to root */
	CHECK(0, wofs_chdir(&wd, "."));
	CHECK(WOFS_ROOT_INODE, wd);
	CHECK(0, wofs_chdir(&wd, ".."));
	CHECK(WOFS_ROOT_INODE, wd);

	/* create a new directory and test it */
	{
		CHECK(0, wofs_mkdir(wd, "Dir1", NULL));
		CHECK(0, wofs_chdir(&wd, "Dir1"));
		CHECK_NE(WOFS_ROOT_INODE, wd);
		const wofs_ino_t new_dir = wd;
		CHECK(0, wofs_chdir(&wd, "."));
		CHECK(new_dir, wd);
		CHECK(0, wofs_chdir(&wd, ".."));
		CHECK(WOFS_ROOT_INODE, wd);

		/* it should be impossible to remove "." and ".." */
		CHECK_NE(0, wofs_rmdir(new_dir, "."));
		CHECK_NE(0, wofs_rmdir(new_dir, ".."));
	}

	/* try illegal names */
	CHECK_NE(0, wofs_mkdir(WOFS_ROOT_INODE, ".", NULL));
	CHECK_NE(0, wofs_mkdir(WOFS_ROOT_INODE, "./", NULL));
	CHECK_NE(0, wofs_mkdir(WOFS_ROOT_INODE, "a/b", NULL));
	CHECK_NE(0, wofs_mkdir(WOFS_ROOT_INODE, extremly_long_name, NULL));

	/* check opendir()/readdir() */
	{
		struct wofs_dir dir;
		CHECK(0, wofs_opendir(&dir, WOFS_ROOT_INODE));
		/* at this point root contains 3 directories: ".", "..", "Dir1" */
		int8_t ec;
		CHECK_NE(NULL, wofs_readdir(&dir, &ec)); CHECK(0, ec);
		CHECK_NE(NULL, wofs_readdir(&dir, &ec)); CHECK(0, ec);
		CHECK_NE(NULL, wofs_readdir(&dir, &ec)); CHECK(0, ec);
		CHECK(NULL, wofs_readdir(&dir, &ec)); CHECK(-WOFS_ERR_ENOMORE, ec);
	}

	/* it should be impossible to remove "." and ".." */
	CHECK_NE(0, wofs_rmdir(WOFS_ROOT_INODE, "."));
	CHECK_NE(0, wofs_rmdir(WOFS_ROOT_INODE, ".."));
	/* and it should not be possible to remove non-existing directories */
	CHECK_NE(0, wofs_rmdir(WOFS_ROOT_INODE, "not-existing"));

	/* remove "Dir1" */
	CHECK(0, wofs_rmdir(WOFS_ROOT_INODE, "Dir1"));
	/* check opendir()/readdir() again, but now we have 1 less entries */
	{
		struct wofs_dir dir;
		CHECK(0, wofs_opendir(&dir, WOFS_ROOT_INODE));
		/* at this point root contains only 2 directories: ".", ".." */
		int8_t ec;
		CHECK_NE(NULL, wofs_readdir(&dir, &ec)); CHECK(0, ec);
		CHECK_NE(NULL, wofs_readdir(&dir, &ec)); CHECK(0, ec);
		CHECK(NULL, wofs_readdir(&dir, &ec)); CHECK(-WOFS_ERR_ENOMORE, ec);
	}

	/* it should be impossible to remove "." and "..", even though root is */
	/* currently empty */
	CHECK_NE(0, wofs_rmdir(WOFS_ROOT_INODE, "."));
	CHECK_NE(0, wofs_rmdir(WOFS_ROOT_INODE, ".."));

	/* check unlink of directories */
	CHECK_NE(0, wofs_unlink(WOFS_ROOT_INODE, "."));
	CHECK_NE(0, wofs_unlink(WOFS_ROOT_INODE, ".."));

	/* check unlink of non-existing files */
	CHECK_NE(0, wofs_unlink(WOFS_ROOT_INODE, "non-existing"));

	/* check open of non-existing files */
	{
		struct wofs_fd fd;
		CHECK_NE(0, wofs_open(&fd, WOFS_ROOT_INODE, "non-existing"));
	}

	/* test creat() */
	{
		struct wofs_fd fd;
		CHECK(0, wofs_creat(&fd, WOFS_ROOT_INODE, "File1"));
		struct wofs_dir dir;
		CHECK(0, wofs_opendir(&dir, WOFS_ROOT_INODE));
		/* at this point root contains 3 entries: ".", "..", "File1" */
		int8_t ec;
		CHECK_NE(NULL, wofs_readdir(&dir, &ec)); CHECK(0, ec);
		CHECK_NE(NULL, wofs_readdir(&dir, &ec)); CHECK(0, ec);
		CHECK_NE(NULL, wofs_readdir(&dir, &ec)); CHECK(0, ec);
		CHECK(NULL, wofs_readdir(&dir, &ec)); CHECK(-WOFS_ERR_ENOMORE, ec);
		CHECK(0, wofs_unlink(WOFS_ROOT_INODE, "File1"));
		/* we shall not be able to unlink it again */
		CHECK_NE(0, wofs_unlink(WOFS_ROOT_INODE, "File1"));
		/* we removed it, we should have 2 entries again */
		CHECK(0, wofs_opendir(&dir, WOFS_ROOT_INODE));
		CHECK_NE(NULL, wofs_readdir(&dir, &ec)); CHECK(0, ec);
		CHECK_NE(NULL, wofs_readdir(&dir, &ec)); CHECK(0, ec);
		CHECK(NULL, wofs_readdir(&dir, &ec)); CHECK(-WOFS_ERR_ENOMORE, ec);
	}

	/* make a directory and add 2 files into it */
	{
		wofs_ino_t dir;
		CHECK(0, wofs_mkdir(WOFS_ROOT_INODE, "Dir2", &dir));
		CHECK_NE(0, wofs_rmdir(WOFS_ROOT_INODE, "./Dir2"));
		struct wofs_fd fd;
		CHECK(0, wofs_creat(&fd, dir, "File2"));
		CHECK(0, wofs_creat(&fd, dir, "File3"));
		/* our directory shall not be removed by now */
		CHECK_NE(0, wofs_rmdir(WOFS_ROOT_INODE, "Dir2"));
	}

	/* test simple write()s, that does not require block allocation */
	{
		struct wofs_fd fd;
		CHECK(0, wofs_creat(&fd, WOFS_ROOT_INODE, "File4"));
		CHECK(6, wofs_write(&fd, "hello ", 6, NULL));
		CHECK(6, wofs_write(&fd, "world!", 6, NULL));
		/* and we can't make a new file with the same name */
		CHECK_NE(0, wofs_creat(&fd, WOFS_ROOT_INODE, "File4"));
		/* but we can reopen it */
		CHECK(0, wofs_open(&fd, WOFS_ROOT_INODE, "File4"));
		char buf[12];
		CHECK(5, wofs_read(&fd, buf, 5, NULL));
		CHECK(7, wofs_read(&fd, buf+5, 7, NULL));
		CHECK(0, memcmp(buf, "hello world!", sizeof(buf)));
	}

	/* test byte-by-byte write() operation */
	{
		struct wofs_fd fd;
		CHECK(0, wofs_creat(&fd, WOFS_ROOT_INODE, "File5"));
		for(wofs_size_t i=0;i<block_size;++i)
		{
			uint8_t byte = (uint8_t)(i & 0xFF);
			CHECK(1, wofs_write(&fd, &byte, 1, NULL));
		}
		/* reopen it */
		CHECK(0, wofs_open(&fd, WOFS_ROOT_INODE, "File5"));
		for(wofs_size_t i=0;i<block_size;++i)
		{
			const uint8_t expected_byte = (uint8_t)(i & 0xFF);
			uint8_t read_byte;
			CHECK(1, wofs_read(&fd, &read_byte, 1, NULL));
			CHECK(0, expected_byte^read_byte);
		}
		CHECK(0, wofs_open(&fd, WOFS_ROOT_INODE, "File5"));
		CHECK(block_size, wofs_seek(&fd, 0, WOFS_SEEK_END, NULL));
	}

	/* test seek() on short file */
	{
		struct wofs_fd fd;
		static const uint8_t data[] = "abcdefghijklmnopqrstuvwxyz";
		uint8_t byte;
		CHECK(0, wofs_creat(&fd, WOFS_ROOT_INODE, "File6"));
		CHECK(26, wofs_write(&fd, data, 26, NULL));
		CHECK(26, wofs_seek(&fd, 0, WOFS_SEEK_CUR, NULL));

		CHECK(25, wofs_seek(&fd, -1, WOFS_SEEK_CUR, NULL));
		CHECK(1, wofs_read(&fd, &byte, 1, NULL));
		CHECK(0, byte ^ data[25]);

		CHECK(20, wofs_seek(&fd, -6, WOFS_SEEK_CUR, NULL));
		CHECK(1, wofs_read(&fd, &byte, 1, NULL));
		CHECK(0, byte ^ data[20]);

		CHECK(0, wofs_seek(&fd, 0, WOFS_SEEK_SET, NULL));
		CHECK(1, wofs_read(&fd, &byte, 1, NULL));
		CHECK(0, byte ^ data[0]);

		CHECK(11, wofs_seek(&fd, 10, WOFS_SEEK_CUR, NULL));
		CHECK(1, wofs_read(&fd, &byte, 1, NULL));
		CHECK(0, byte ^ data[11]);

		CHECK(25, wofs_seek(&fd, -1, WOFS_SEEK_END, NULL));
		CHECK(1, wofs_read(&fd, &byte, 1, NULL));
		CHECK(0, byte ^ data[25]);

		CHECK(26, wofs_seek(&fd, 0, WOFS_SEEK_CUR, NULL));
	}

	/* test seek() on a multi-block file */
	{
		struct wofs_fd fd;
		CHECK(0, wofs_creat(&fd, WOFS_ROOT_INODE, "File7"));
		wofs_size_t written_bytes;
		for(written_bytes=0;
			written_bytes<=(block_size*3);
			written_bytes+=sizeof(all_bytes))
		{
			CHECK(sizeof(all_bytes),
				wofs_write(&fd, all_bytes, sizeof(all_bytes), NULL));
		}

		const wofs_off_t size = wofs_seek(&fd, 0, WOFS_SEEK_CUR, NULL);
		CHECK(written_bytes, size);

		uint8_t byte;

		wofs_off_t off;

		off = block_size/2;
		CHECK(off, wofs_seek(&fd, off, WOFS_SEEK_SET, NULL));
		CHECK(1, wofs_read(&fd, &byte, 1, NULL));
		CHECK(0, byte ^ all_bytes[off % sizeof(all_bytes)]);

		off = block_size;
		CHECK(off, wofs_seek(&fd, off, WOFS_SEEK_SET, NULL));
		CHECK(1, wofs_read(&fd, &byte, 1, NULL));
		CHECK(0, byte ^ all_bytes[off % sizeof(all_bytes)]);

		off = block_size/2;
		CHECK(off, wofs_seek(&fd, off, WOFS_SEEK_SET, NULL));
		CHECK(1, wofs_read(&fd, &byte, 1, NULL));
		CHECK(0, byte ^ all_bytes[off % sizeof(all_bytes)]);

		off = (block_size*7)/3;
		CHECK(off, wofs_seek(&fd, off, WOFS_SEEK_SET, NULL));
		CHECK(1, wofs_read(&fd, &byte, 1, NULL));
		CHECK(0, byte ^ all_bytes[off % sizeof(all_bytes)]);

		off = (block_size*7)/9;
		CHECK(off, wofs_seek(&fd, off, WOFS_SEEK_SET, NULL));
		CHECK(1, wofs_read(&fd, &byte, 1, NULL));
		CHECK(0, byte ^ all_bytes[off % sizeof(all_bytes)]);

		off = block_size*2;
		CHECK(off, wofs_seek(&fd, off, WOFS_SEEK_SET, NULL));
		CHECK(1, wofs_read(&fd, &byte, 1, NULL));
		CHECK(0, byte ^ all_bytes[off % sizeof(all_bytes)]);

		off = block_size*2 + 1;
		CHECK(off, wofs_seek(&fd, off, WOFS_SEEK_SET, NULL));
		CHECK(1, wofs_read(&fd, &byte, 1, NULL));
		CHECK(0, byte ^ all_bytes[off % sizeof(all_bytes)]);

		off = block_size*2 + 1; /* once again! */
		CHECK(off, wofs_seek(&fd, off, WOFS_SEEK_SET, NULL));
		CHECK(1, wofs_read(&fd, &byte, 1, NULL));
		CHECK(0, byte ^ all_bytes[off % sizeof(all_bytes)]);

		off = block_size*2 + 5;
		CHECK(off, wofs_seek(&fd, off, WOFS_SEEK_SET, NULL));
		CHECK(1, wofs_read(&fd, &byte, 1, NULL));
		CHECK(0, byte ^ all_bytes[off % sizeof(all_bytes)]);

		off = block_size/3;
		CHECK(off, wofs_seek(&fd, off, WOFS_SEEK_SET, NULL));
		CHECK(1, wofs_read(&fd, &byte, 1, NULL));
		CHECK(0, byte ^ all_bytes[off % sizeof(all_bytes)]);

		off = block_size*3;
		CHECK(off, wofs_seek(&fd, off, WOFS_SEEK_SET, NULL));
		CHECK(1, wofs_read(&fd, &byte, 1, NULL));
		CHECK(0, byte ^ all_bytes[off % sizeof(all_bytes)]);

		{
			int8_t ec;
			off = wofs_seek(&fd, -2, WOFS_SEEK_END, &ec);
			CHECK(0, ec);
			CHECK(1, wofs_read(&fd, &byte, 1, NULL));
			CHECK(0, byte ^ all_bytes[off % sizeof(all_bytes)]);
			CHECK(off+2, wofs_seek(&fd, 0, WOFS_SEEK_END, NULL));
		}

		CHECK(0, wofs_unlink(WOFS_ROOT_INODE, "File7"));
		/* we shall not be able to unlink it again */
		CHECK_NE(0, wofs_unlink(WOFS_ROOT_INODE, "File7"));
	}

	/* test fallocate() */
	{
		struct wofs_fd fd;
		CHECK(0, wofs_creat(&fd, WOFS_ROOT_INODE, "File8"));
		CHECK(0, wofs_fallocate(&fd, 20));
		CHECK(0, wofs_seek(&fd, 0, WOFS_SEEK_SET, NULL));
		CHECK(20, wofs_seek(&fd, 0, WOFS_SEEK_END, NULL));
		CHECK(0, wofs_fallocate(&fd, 10));
		CHECK(0, wofs_seek(&fd, 0, WOFS_SEEK_SET, NULL));
		CHECK(20, wofs_seek(&fd, 0, WOFS_SEEK_END, NULL));
		CHECK(0, wofs_fallocate(&fd, block_size));
		CHECK(0, wofs_seek(&fd, 0, WOFS_SEEK_SET, NULL));
		CHECK(block_size, wofs_seek(&fd, 0, WOFS_SEEK_END, NULL));
		CHECK(0, wofs_fallocate(&fd, 0)); /* bogus length */
		CHECK(0, wofs_fallocate(&fd, block_size/2)); /* bogus length */
		CHECK(0, wofs_seek(&fd, 0, WOFS_SEEK_SET, NULL));
		CHECK(block_size, wofs_seek(&fd, 0, WOFS_SEEK_END, NULL));

		/* byebye */
		CHECK(0, wofs_unlink(WOFS_ROOT_INODE, "File8"));
		/* we shall not be able to unlink it again */
		CHECK_NE(0, wofs_unlink(WOFS_ROOT_INODE, "File8"));
	}

	/* test istat() */
	{
		uint8_t type;
		CHECK(0, wofs_istat(WOFS_ROOT_INODE, &type));
		CHECK(WOFS_INOTYPE_DIR, type);
	}

	{
		struct wofs_statfs st1;

		CHECK(0, wofs_statfs(&st1));

		CHECK(0, wofs_mount(nblocks, blkszsh, 0));

		struct wofs_statfs st2;
		CHECK(0, wofs_statfs(&st2));

		CHECK(st1.f_blocks, st2.f_blocks);
		CHECK(st1.f_bfree, st2.f_bfree);
	}

	/* fill at least one directory with entries, and remove them */
	{
		wofs_ino_t dir;

		CHECK(0, wofs_mkdir(WOFS_ROOT_INODE, "Dir3", &dir));

		for(wofs_size_t i=0;i<block_size;i+=WRONCEFS_NAME_MAX)
		{
			char name[WRONCEFS_NAME_MAX+1];
			sprintf(name, "%*.*u", (int)WRONCEFS_NAME_MAX,
				(int)WRONCEFS_NAME_MAX, i);

			CHECK(0, wofs_mkdir(dir, name, NULL));
		}

		for(wofs_size_t i=0;i<block_size;i+=WRONCEFS_NAME_MAX)
		{
			char name[WRONCEFS_NAME_MAX+1];
			sprintf(name, "%*.*u", (int)WRONCEFS_NAME_MAX,
				(int)WRONCEFS_NAME_MAX, i);

			CHECK(0, wofs_rmdir(dir, name));
		}

		CHECK(0, wofs_rmdir(WOFS_ROOT_INODE, "Dir3"));
	}

	/* test various mount fixups */
	{
		wofs_ino_t dir;

		for(wofs_size_t i=1;i<=block_size;++i)
		{
			/* remount */
			CHECK(0, wofs_mount(nblocks, blkszsh, WOFS_MOUNT_NORMAL));
			CHECK(0, wofs_fsck(nblocks, blkszsh, NULL));

			/* this might exist, might not */
			wofs_rmdir(WOFS_ROOT_INODE, "Dir4");

			io_die_after(i);
			/* this might not work due to the write limiter */
			wofs_mkdir(WOFS_ROOT_INODE, "Dir4", NULL);
			io_die_reset();
		}

		/* cleanup */
		wofs_rmdir(WOFS_ROOT_INODE, "Dir4");

		/* now, everything should back on track */
		CHECK(0, wofs_mkdir(WOFS_ROOT_INODE, "Dir5", NULL));
	}

	/* test fallocate() with a larger size */
	{
		struct wofs_fd fd;
		CHECK(0, wofs_creat(&fd, WOFS_ROOT_INODE, "File9"));
		CHECK(0, wofs_seek(&fd, 0, WOFS_SEEK_END, NULL));
		CHECK(0, wofs_fallocate(&fd, (3*block_size)/2));
		CHECK(0, wofs_open(&fd, WOFS_ROOT_INODE, "File9"));
		CHECK((3*block_size)/2, wofs_seek(&fd, 0, WOFS_SEEK_END, NULL));
	}

	/* test file write when interrupted at various places */
	{
		wofs_ino_t dir;
		struct wofs_fd fd;

		CHECK(0, wofs_mkdir(WOFS_ROOT_INODE, "Dir6", &dir));

		for(wofs_size_t i=1;i<=block_size*2;++i)
		{
			/* remount */
			CHECK(0, wofs_fsck(nblocks, blkszsh, NULL));

			/* this might exist, might not */
			wofs_unlink(dir, "File10");

			CHECK(0, wofs_creat(&fd, dir, "File10"));

			io_die_after(i);

			wofs_size_t written_bytes;
			for(written_bytes=0;written_bytes<block_size;)
			{
				wofs_size_t part = block_size - written_bytes;
				if(part>sizeof(all_bytes)) { part = sizeof(all_bytes); }

				/* this might not work due to the write limiter */
				wofs_write(&fd, all_bytes, part, NULL);

				written_bytes += part;
			}

			wofs_unlink(dir, "File10");

			io_die_reset();
		}
	}

	return 0;
}
